print("import: 'prompt_toolkit'")
import prompt_toolkit

print("import: 'prompt_toolkit.clipboard'")
import prompt_toolkit.clipboard

print("import: 'prompt_toolkit.contrib'")
import prompt_toolkit.contrib

print("import: 'prompt_toolkit.contrib.completers'")
import prompt_toolkit.contrib.completers

print("import: 'prompt_toolkit.contrib.regular_languages'")
import prompt_toolkit.contrib.regular_languages

print("import: 'prompt_toolkit.contrib.telnet'")
import prompt_toolkit.contrib.telnet

print("import: 'prompt_toolkit.eventloop'")
import prompt_toolkit.eventloop

print("import: 'prompt_toolkit.filters'")
import prompt_toolkit.filters

print("import: 'prompt_toolkit.key_binding'")
import prompt_toolkit.key_binding

print("import: 'prompt_toolkit.key_binding.bindings'")
import prompt_toolkit.key_binding.bindings

print("import: 'prompt_toolkit.layout'")
import prompt_toolkit.layout

